// This program doesn't support Unicode.

// for Win98 or later.
#define _WIN32_WINDOWS 0x0410

#include <windows.h>
#include "runtime.h"
#include "resource.h"
#include "bit.h"

#define MIDIMSG2(stat, data1) (DWORD)((stat) | ((data1) << 8))
#define MIDIMSG3(stat, data1, data2) (DWORD)((stat) | ((data1) << 8) | ((data2) << 16))

typedef struct
{
	BYTE* data;
	DWORD size;
}CHUNK;

#define SwapDword(data) (*((BYTE*)(data) + 0) << 24 | *((BYTE*)(data) + 1) << 16 | *((BYTE*)(data) + 2) << 8 | *((BYTE*)(data) + 3))
#define SwapWord(data) (*((BYTE*)(data) + 0) << 8 | *((BYTE*)(data) + 1))

typedef struct _tag_EVENTLIST
{
	DWORD duration;
	DWORD time;
	BYTE status;
	BYTE* data;
	DWORD size;
	struct _tag_EVENTLIST* next;
}EVENT;

EVENT* eventlist = NULL;
EVENT* current;

typedef struct _tag_NOTEOFF
{
	DWORD ch;
	DWORD note;
	DWORD end;
	struct _tag_NOTEOFF* next;
}NOTEOFF;

NOTEOFF* noteofflist = NULL;

typedef struct
{
	DWORD dbase;	// Duration TimeBase
	DWORD gbase;	// Gatetime TimeBase
	EVENT* events;
}MODULE;

static BYTE* smaf = NULL;
static DWORD smafsize;
static DWORD dbase;
static DWORD gbase;

BOOL CALLBACK MessageDlgWndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	if(msg == WM_INITDIALOG)
		return TRUE;
	if(msg == WM_COMMAND && (wParam == IDOK || wParam == IDCANCEL))
		EndDialog(hWnd, wParam);

	return FALSE;
}

BOOL findchunk4(BYTE* data, DWORD dsize, char* cname, CHUNK* ch)
{
	DWORD cur;

	for(cur = 0; cur < dsize; )
	{
		if(!memcomp(data + cur, cname, 4))
		{
			ch->data = data + cur + 8;
			ch->size = SwapDword(data + cur + 4);
			return TRUE;
		}
		cur += 8 + SwapDword(data + cur + 4);
	}

	return FALSE;
}

BOOL findchunk3(BYTE* data, DWORD dsize, char* cname, CHUNK* ch)
{
	DWORD cur;

	for(cur = 0; cur < dsize; )
	{
		if(!memcomp(data + cur, cname, 3))
		{
			ch->data = data + cur + 8;
			ch->size = SwapDword(data + cur + 4);
			return TRUE;
		}
		cur += 8 + SwapDword(data + cur + 4);
	}

	return FALSE;
}

DWORD GetVValAndForward(BYTE** pdata)
{
	DWORD val = 0;

	do
	{
		val <<= 7;
		val |= **pdata & 0x7F;
	}while(*(*pdata)++ & 0x80);

	return val;
}

DWORD GetVVal(BYTE* pdata)
{
	DWORD val = 0;

	do
	{
		val <<= 7;
		val |= *pdata & 0x7F;
	}while(*pdata++ & 0x80);

	return val;
}

void ClearEvents(void)
{
	EVENT* e = eventlist;

	while(e)
	{
		EVENT* next = e->next;
		hfree(e->data);
		hfree(e);
		e = next;
	}

	eventlist = NULL;

	return;
}

// for HPS: convert to MSF
void ConvertEvents2(BYTE* data, DWORD size, BOOL fsetup)
{
	EVENT* ev = eventlist;
	BYTE* p;
	DWORD tdur = 0;
	signed int octshift = 0;

	if(ev)
		while(ev->next)
			ev = ev->next;

	for(p = data; p < data + size; )
	{	// Ignore error ;-(
		// Maybe, there is more nice efficiency routine.
		EVENT* e;
		BYTE esize = 1;

		e = halloc(sizeof(EVENT));
		if(fsetup)
			e->duration = 0;
		else
			e->duration = GetVValAndForward(&p);
		tdur += e->duration * dbase;
		e->time = tdur;
		e->status = 0;
		e->size = 0;
		e->data = NULL;

		if(p[0] == 0xFF && p[1] == 0xF0)	// Exclusive
		{
			esize = 2 + p[2] + 1;
			e->size = p[2] + 1;
			e->status = 0xF0;
			e->data = halloc(e->size);
			memcopy(e->data, p + 2, e->size);
		}else if(p[0] == 0xFF && p[1] == 0x00)	// NOP
		{
			esize = 2;
			e->size = 1;
			e->data = halloc(e->size);
			e->status = 0xFF;
			e->data[0] = 0x00;
		}else
		{
			if(p[0] != 0x00)	// Note
			{
				BYTE ch;
				BYTE oct;
				BYTE note;
				BYTE absnote;

				ch = p[0] >> 6 & 0x03;
				oct = p[0] >> 4 & 0x03;
				note = p[0] & 0x0F;
				// Midi-69 = 440Hz(A) = Oct2,9
				absnote = 36 + (oct + octshift) * 12 + note;

				e->status = 0x80 | ch;
				{
					BYTE* q = p + 1;

					GetVValAndForward(&q);
					esize = q - p;
					e->size = q - p;
					e->data = halloc(e->size);

					e->data[0] = absnote;
					if(p[1] & 0x80)
					{
						e->data[1] = p[1];
						e->data[2] = p[2];
					}else
						e->data[1] = p[1];
				}
			}else
			{
				BYTE ch;
				ch = p[1] >> 6 & 0x03;

				if(p[1] == 0x00 && p[2] == 0x00 && p[3] == 0x00)	// EOS?
					esize = 4;
				if((p[1] & 0x30) == 0x30)
				{
					if((p[1] & 0x0F) == 0x00)	// Program Change
					{
						esize = 3;
						e->size = 1;
						e->data = halloc(e->size);
						e->status = 0xC0 | ch;
						e->data[0] = p[2];
					}
					if((p[1] & 0x0F) == 0x01)	// Bank Select
					{
						esize = 3;
						e->size = 2;
						e->data = halloc(e->size);
						e->status = 0xB0 | ch;
						e->data[0] = 0x00;	// banksel
						e->data[1] = p[2];
					}
					if((p[1] & 0x0F) == 0x02)	// Octave Shift
					{
						esize = 3;
						octshift = (signed char)p[2];

						// NOP
						e->size = 1;
						e->data = halloc(e->size);
						e->status = 0xFF;
						e->data[0] = 0x00;
					}
					if((p[1] & 0x0F) == 0x03)	// Modulation
					{
						esize = 3;
						e->size = 2;
						e->data = halloc(e->size);
						e->status = 0xB0 | ch;
						e->data[0] = 0x01;	// mod.
						e->data[1] = p[2];
					}
					if((p[1] & 0x0F) == 0x04)	// PitchBend
					{
						esize = 3;
						e->size = 2;
						e->data = halloc(e->size);
						e->status = 0xE0 | ch;
						e->data[0] = 0x00;
						e->data[1] = p[2];	// tenuki.
					}
					if((p[1] & 0x0F) == 0x07)	// Volume
					{
						esize = 3;
						e->size = 2;
						e->data = halloc(e->size);
						e->status = 0xB0 | ch;
						e->data[0] = 0x07;	// vol
						e->data[1] = p[2];
					}
					if((p[1] & 0x0F) == 0x0A)	// Pan
					{
						esize = 3;
						e->size = 2;
						e->data = halloc(e->size);
						e->status = 0xB0 | ch;
						e->data[0] = 0x0A;	// pan
						e->data[1] = p[2];
					}
					if((p[1] & 0x0F) == 0x0B)	// Expression
					{
						esize = 3;
						e->size = 2;
						e->data = halloc(e->size);
						e->status = 0xB0 | ch;
						e->data[0] = 0x0B;	// exp.
						e->data[1] = p[2];
					}
				}else
				{
					if((p[1] & 0x30) == 0x00)	// Expression
					{
						esize = 2;
						e->size = 2;
						e->data = halloc(e->size);
						e->status = 0xB0 | ch;
						e->data[0] = 0x0B;	// exp.
						e->data[1] = (p[1] & 0xF0) << 3;
					}
					if((p[1] & 0x30) == 0x10)	// PitchBend
					{
						esize = 2;
						e->size = 2;
						e->data = halloc(e->size);
						e->status = 0xE0 | ch;
						e->data[0] = 0x00;
						e->data[1] = (p[1] & 0xF0) << 3;	// chou tenuki
					}
					if((p[1] & 0x30) == 0x20)	// Modulation
					{
						esize = 2;
						e->size = 2;
						e->data = halloc(e->size);
						e->status = 0xB0 | ch;
						e->data[0] = 0x01;	// mod.
						e->data[1] = (p[1] & 0xF0) << 3;
					}
				}
			}
		}

		if(esize == 1)
		{
			TCHAR ts[256];

			wsprintf(ts, "What's this? %02X", p[0]);
			MessageBox(NULL, ts, "Error", MB_ICONERROR | MB_OK);
		}

		p += esize;

		e->next = NULL;

		if(!ev)
			eventlist = e;
		else
			ev->next = e;

		ev = e;
	}

	return;
}

// for MSF
void ConvertEvents(BYTE* data, DWORD size, BOOL fsetup)
{
	EVENT* ev = eventlist;
	BYTE* p;
	DWORD tdur = 0;

	if(ev)
		while(ev->next)
			ev = ev->next;

	for(p = data; p < data + size; )
	{	// Ignore error ;-(
		// Maybe, there is more nice efficiency routine.
		EVENT* e;

		e = halloc(sizeof(EVENT));
		if(fsetup)
			e->duration = 0;
		else
			e->duration = GetVValAndForward(&p);
		tdur += e->duration * dbase;
		e->time = tdur;
		e->status = *p++;
		e->size = 0;
		e->data = NULL;

		if((e->status & 0xF0) == 0x80)	// Note w/o verocity
		{
			BYTE* q = p + 1;

			GetVValAndForward(&q);
			e->size = q - p;
		}
		if((e->status & 0xF0) == 0x90)	// Note w/ verocity
		{
			BYTE* q = p + 2;

			GetVValAndForward(&q);
			e->size = q - p;
		}
		if((e->status & 0xF0) == 0xA0)	// Resv 3byte
		{
			e->size = 3;
		}
		if((e->status & 0xF0) == 0xB0)	// Control Change
		{
			e->size = 2;
		}
		if((e->status & 0xF0) == 0xC0)	// Program Change
		{
			e->size = 1;
		}
		if((e->status & 0xF0) == 0xD0)	// Resv 2byte
		{
			e->size = 2;
		}
		if((e->status & 0xF0) == 0xE0)	// PitchBend
		{
			e->size = 2;
		}
		if(e->status == 0xF0)	// Exclusive
		{
			BYTE* q = p;

			e->size = GetVValAndForward(&q);
			e->size += (q - p);	// VVal size
		}
		if(e->status == 0xFF)	// NOP or EOS
		{
			if(*p == 0x00)
				e->size = 1;
			if(*p == 0x2F)
				e->size = 2;
		}

		// copy.
		if(e->size)
		{
			e->data = halloc(e->size);
			memcopy(e->data, p, e->size);
			p += e->size;
		}else
			MessageBox(NULL, "Bug or illegal", "What", MB_OK);

		e->next = NULL;

		if(!ev)
			eventlist = e;
		else
			ev->next = e;

		ev = e;
	}

	return;
}

void draw(HWND hWnd, HDC hdc, int width, int height)
{
	HFONT hfold;

	SetTextColor(hdc, RGB(0, 0, 0));
	SetBkMode(hdc, TRANSPARENT);

	if(!smaf)
	{
		FillRectangle(hdc, 0, 0, width, height, GetSysColor(COLOR_3DFACE));
		MoveToEx(hdc, 0, 0, NULL);
		PutText(hdc, "This is output window");

		return;
	}

	FillRectangle(hdc, 0, 0, width, height, RGB(220, 220, 220));

	{
		LOGFONT lf = {0};

		lf.lfCharSet = DEFAULT_CHARSET;
		strcopy(lf.lfFaceName, "lr SVbN");
		lf.lfHeight = -12;

		hfold = SelectObject(hdc, CreateFontIndirect(&lf));
	}

	{
		SCROLLINFO vsi, hsi;
		SIZE s;

		vsi.cbSize = hsi.cbSize = sizeof(SCROLLINFO);
		vsi.fMask = hsi.fMask = SIF_POS;
		GetScrollInfo(hWnd, SB_VERT, &vsi);
		GetScrollInfo(hWnd, SB_HORZ, &hsi);

		GetTextExtentPoint32(hdc, "A", 1, &s);
		MoveToEx(hdc, hsi.nPos * -s.cx, vsi.nPos * -s.cy, NULL);
	}

	{
		CHUNK mmmd;
		CHUNK dummy;

		if(!findchunk4(smaf, smafsize, "MMMD", &mmmd))
		{
			PutText(hdc, "MMMD NOT FOUND");
			return;
		}

		mmmd.size -= 2;	// CRC

		{
			CHUNK cnti;

			if(!findchunk4(mmmd.data, mmmd.size, "CNTI", &cnti))
			{
				PutText(hdc, "CNTI NOT FOUND");
				return;
			}

			PutText(hdc, "CNTI info");

			if(cnti.data[0] == 0x00)
				PutText(hdc, "Contents class: YAMAHA");
			else
				PutText(hdc, "Contents class: Not yamaha %02X", cnti.data[0]);

			if((0x00 <= cnti.data[1] && cnti.data[1] <= 0x0F) || (0x30 <= cnti.data[1] && cnti.data[1] <= 0x33))
				PutText(hdc, "Contents type: Ringtone %02X", cnti.data[1]);
			else if((0x10 <= cnti.data[1] && cnti.data[1] <= 0x1F) || (0x40 <= cnti.data[1] && cnti.data[1] <= 0x42))
				PutText(hdc, "Contents type: Karaoke %02X", cnti.data[1]);
			else if((0x20 <= cnti.data[1] && cnti.data[1] <= 0x2F) || (0x50 <= cnti.data[1] && cnti.data[1] <= 0x53))
				PutText(hdc, "Contents type: CM %02X", cnti.data[1]);
			else
				PutText(hdc, "Contents type: RESERVED %02X", cnti.data[1]);

			switch(cnti.data[2])
			{
			case 0x00: PutText(hdc, "Contents code: Shift-JIS"); break;
			case 0x01: PutText(hdc, "Contents code: Latin-1"); break;
			case 0x02: PutText(hdc, "Contents code: EUC-KR"); break;
			case 0x03: PutText(hdc, "Contents code: GB-2312"); break;
			case 0x04: PutText(hdc, "Contents code: Big5"); break;
			case 0x05: PutText(hdc, "Contents code: KOI8-R"); break;
			case 0x06: PutText(hdc, "Contents code: TCVN-5773:1993"); break;
			case 0x20: PutText(hdc, "Contents code: UCS-2"); break;
			case 0x21: PutText(hdc, "Contents code: UCS-4"); break;
			case 0x22: PutText(hdc, "Contents code: UTF-7"); break;
			case 0x23: PutText(hdc, "Contents code: UTF-8"); break;
			case 0x24: PutText(hdc, "Contents code: UTF-16"); break;
			case 0x25: PutText(hdc, "Contents code: UTF-32"); break;
			default: PutText(hdc, "Contents code: RESERVED %02X", cnti.data[2]); break;
			}

			if(cnti.data[3] & 0x04)
				PutText(hdc, "CopyStatus: Edit NG");
			else
				PutText(hdc, "CopyStatus: Edit OK");
			if(cnti.data[3] & 0x02)
				PutText(hdc, "CopyStatus: Save NG");
			else
				PutText(hdc, "CopyStatus: Save OK");
			if(cnti.data[3] & 0x01)
				PutText(hdc, "CopyStatus: Trans NG");
			else
				PutText(hdc, "CopyStatus: Trans OK");

			PutText(hdc, "Copy Count: %d", cnti.data[4]);

			if(cnti.size > 5)
			{
				DWORD i, j;

				PutText(hdc, "Optional data followed.");

				for(i = 5; i < cnti.size; )
				{
					char tag[16];
					char data[256];

					tag[0] = cnti.data[i];
					tag[1] = cnti.data[i + 1];
					tag[2] = '\0';

					if(cnti.data[i] == 'V' && cnti.data[i + 1] == 'N')
						lstrcpy(tag, "Vendor");
					if(cnti.data[i] == 'C' && cnti.data[i + 1] == 'N')
						lstrcpy(tag, "Carrior");
					if(cnti.data[i] == 'C' && cnti.data[i + 1] == 'A')
						lstrcpy(tag, "Category");
					if(cnti.data[i] == 'S' && cnti.data[i + 1] == 'T')
						lstrcpy(tag, "Title");
					if(cnti.data[i] == 'A' && cnti.data[i + 1] == 'N')
						lstrcpy(tag, "Artist");
					if(cnti.data[i] == 'W' && cnti.data[i + 1] == 'W')
						lstrcpy(tag, "Word");
					if(cnti.data[i] == 'S' && cnti.data[i + 1] == 'W')
						lstrcpy(tag, "Music");
					if(cnti.data[i] == 'A' && cnti.data[i + 1] == 'W')
						lstrcpy(tag, "Edit");
					if(cnti.data[i] == 'C' && cnti.data[i + 1] == 'R')
						lstrcpy(tag, "Copyright");
					if(cnti.data[i] == 'G' && cnti.data[i + 1] == 'R')
						lstrcpy(tag, "CR-Group");
					if(cnti.data[i] == 'M' && cnti.data[i + 1] == 'I')
						lstrcpy(tag, "ManInfo");
					if(cnti.data[i] == 'C' && cnti.data[i + 1] == 'D')
						lstrcpy(tag, "Create");
					if(cnti.data[i] == 'U' && cnti.data[i + 1] == 'D')
						lstrcpy(tag, "Update");
					if(cnti.data[i] == ' ' && cnti.data[i + 1] == 'C')
						lstrcpy(tag, "Copy by?");
					i += 3;

					for(j = i; /*Fact=IllegalFile->*/j < cnti.size && cnti.data[j] != ','; j++)
						;
					memcopy(data, cnti.data + i, j - i);
					data[j - i] = '\0';

					i = j + 1;

					{
						TCHAR ts[256];
						wsprintf(ts, "%s: %s", tag, data);
						PutText(hdc, ts);
					}
				}
			}
		}

		{
			CHUNK opda;

			if(findchunk4(mmmd.data, mmmd.size, "OPDA", &opda))
			{
				CHUNK dch;
				PutText(hdc, "OPDA found.");

				if(findchunk3(opda.data, opda.size, "Dch", &dch))
				{
					DWORD i;
					PutText(hdc, "Dch found.");

					for(i = 0; i < dch.size; )
					{
						char tag[16];
						char data[256];

						tag[0] = dch.data[i];
						tag[1] = dch.data[i + 1];
						tag[2] = '\0';

						if(dch.data[i] == 'V' && dch.data[i + 1] == 'N')
							lstrcpy(tag, "Vendor");
						if(dch.data[i] == 'C' && dch.data[i + 1] == 'N')
							lstrcpy(tag, "Carrior");
						if(dch.data[i] == 'C' && dch.data[i + 1] == 'A')
							lstrcpy(tag, "Category");
						if(dch.data[i] == 'S' && dch.data[i + 1] == 'T')
							lstrcpy(tag, "Title");
						if(dch.data[i] == 'A' && dch.data[i + 1] == 'N')
							lstrcpy(tag, "Artist");
						if(dch.data[i] == 'W' && dch.data[i + 1] == 'W')
							lstrcpy(tag, "Word");
						if(dch.data[i] == 'S' && dch.data[i + 1] == 'W')
							lstrcpy(tag, "Music");
						if(dch.data[i] == 'A' && dch.data[i + 1] == 'W')
							lstrcpy(tag, "Edit");
						if(dch.data[i] == 'C' && dch.data[i + 1] == 'R')
							lstrcpy(tag, "Copyright");
						if(dch.data[i] == 'G' && dch.data[i + 1] == 'R')
							lstrcpy(tag, "CR-Group");
						if(dch.data[i] == 'M' && dch.data[i + 1] == 'I')
							lstrcpy(tag, "ManInfo");
						if(dch.data[i] == 'C' && dch.data[i + 1] == 'D')
							lstrcpy(tag, "Create");
						if(dch.data[i] == 'U' && dch.data[i + 1] == 'D')
							lstrcpy(tag, "Update");
						if(dch.data[i] == 'E' && dch.data[i + 1] == 'S')
							lstrcpy(tag, "EditState");
						if(dch.data[i] == 'V' && dch.data[i + 1] == 'C')
							lstrcpy(tag, "vCard");
						if(dch.data[i] == ' ' && dch.data[i + 1] == 'C')
							lstrcpy(tag, "Copy by?");
						i += 2;

						memcopy(data, dch.data + i + 2, SwapWord(dch.data + i));
						data[SwapWord(dch.data + i)] = '\0';

						i += SwapWord(dch.data + i) + 2;

						{
							TCHAR ts[256];
							wsprintf(ts, "%s: %s", tag, data);
							PutText(hdc, ts);
						}					
					
					}
				}
			}
		}

		{
			CHUNK mtr;

			if(findchunk3(mmmd.data, mmmd.size, "MTR", &mtr))
			{
				PutText(hdc, "MTR info");

				switch(mtr.data[0])
				{
				case 0x00: PutText(hdc, "Format Type: HandyPhone Standard (MA-1,2)"); break;
				case 0x01: PutText(hdc, "Format Type: Mobile Standard(Compress)"); break;
				case 0x02: PutText(hdc, "Format Type: Mobile Standard(No Compress) (MA-3,5)"); break;
				case 0x03: PutText(hdc, "Format Type: MA-7"); break;
				default: PutText(hdc, "Format Type: RESERVED %02X", mtr.data[0]); break;
				}

				switch(mtr.data[1])
				{
				case 0x00: PutText(hdc, "Sequence Type: Stream sequence"); break;
				case 0x01: PutText(hdc, "Sequence Type: Sub-sequence"); break;
				default: PutText(hdc, "Sequence Type: RESERVED %02X", mtr.data[0]); break;
				}

				switch(mtr.data[2])
				{
				case 0x00: PutText(hdc, "TimeBase(Duration): 1 msec"); break;
				case 0x01: PutText(hdc, "TimeBase(Duration): 2 msec"); break;
				case 0x02: PutText(hdc, "TimeBase(Duration): 4 msec"); break;
				case 0x03: PutText(hdc, "TimeBase(Duration): 5 msec"); break;
				case 0x10: PutText(hdc, "TimeBase(Duration): 10 msec"); break;
				case 0x11: PutText(hdc, "TimeBase(Duration): 20 msec"); break;
				case 0x12: PutText(hdc, "TimeBase(Duration): 40 msec"); break;
				case 0x13: PutText(hdc, "TimeBase(Duration): 50 msec"); break;
				default: PutText(hdc, "Sequence Type: RESERVED %02X", mtr.data[0]); break;
				}

				switch(mtr.data[3])
				{
				case 0x00: PutText(hdc, "TimeBase(GateTime): 1 msec"); break;
				case 0x01: PutText(hdc, "TimeBase(GateTime): 2 msec"); break;
				case 0x02: PutText(hdc, "TimeBase(GateTime): 4 msec"); break;
				case 0x03: PutText(hdc, "TimeBase(GateTime): 5 msec"); break;
				case 0x10: PutText(hdc, "TimeBase(GateTime): 10 msec"); break;
				case 0x11: PutText(hdc, "TimeBase(GateTime): 20 msec"); break;
				case 0x12: PutText(hdc, "TimeBase(GateTime): 40 msec"); break;
				case 0x13: PutText(hdc, "TimeBase(GateTime): 50 msec"); break;
				default: PutText(hdc, "Sequence Type: RESERVED %02X", mtr.data[0]); break;
				}

				if(mtr.data[0] == 0x02)
				{
					PutText(hdc, "Analysing MA-3,5 deep...");

					{
						int i;
						for(i = 0; i < 16; i++)
						{
							TCHAR ts[1024];
							wsprintf(ts, "Channel %i: KCS=", i);
							switch(mtr.data[4 + i] >> 6)
							{
							case 0x00: lstrcat(ts, "None "); break;
							case 0x01: lstrcat(ts, "OFF "); break;
							case 0x02: lstrcat(ts, "ON "); break;
							case 0x03: lstrcat(ts, "RESERVED "); break;
							}
							if(mtr.data[4 + i] & 0x20)
								lstrcat(ts, "Vib ");
							if(mtr.data[4 + i] & 0x10)
								lstrcat(ts, "LED ");
							switch(mtr.data[4 + i] & 0x03)
							{
							case 0x00: lstrcat(ts, "Nocare "); break;
							case 0x01: lstrcat(ts, "Melody "); break;
							case 0x02: lstrcat(ts, "NON Melody "); break;
							case 0x03: lstrcat(ts, "Rhythm "); break;
							}
							PutText(hdc, ts);
						}
					}

					mtr.data += 4 + 16;

					if(findchunk4(mtr.data, mtr.size, "MspI", &dummy))
						PutText(hdc, "MspI Info found.");

					if(findchunk4(mtr.data, mtr.size, "Mtsu", &dummy))
						PutText(hdc, "Mtsu Setup found.");

					if(findchunk4(mtr.data, mtr.size, "Mtsq", &dummy))
						PutText(hdc, "Mtsq Sequence found.");

					if(findchunk4(mtr.data, mtr.size, "Mtsp", &dummy))
						PutText(hdc, "Mtsp PCM found.");
				}
			}
		}
	}

	PutText(hdc, "EventList!");

	{
		EVENT* e = eventlist;
		DWORD count = 0;

		PutText(hdc, "<Line No.> Dur. St(Len): data...");
		PutText(hdc, "---");

		while(e)
		{
			TCHAR ts[1024];
			DWORD i;

			wsprintf(ts, "<%08d> %08dms %02X(%3d): ", count++, e->time, e->status, e->size);
			for(i = 0; i < e->size; i++)
				wsprintf(ts, "%s%02X ", ts, e->data[i]);
			PutText(hdc, ts);
			e = e->next;
		}
	}

	PutText(hdc, "READY");

	DeleteObject(SelectObject(hdc, hfold));

	return;
}

void FreeSmaf(void)
{
	NOTEOFF* no = noteofflist;

	if(smaf)
		hfree(smaf);

	ClearEvents();

	while(no)
	{
		NOTEOFF* nx = no->next;
		hfree(no);
		no = nx;
	}
	noteofflist = NULL;

	smaf = NULL;
}

void LoadSmaf(LPCTSTR fname)
{
	HANDLE hf = NULL;
	DWORD read;

	if(smaf)
		FreeSmaf();

	if((hf = CreateFile(fname, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL)) == INVALID_HANDLE_VALUE)
	{
		TCHAR ts[MAX_PATH + 80];

		wsprintf(ts, "Can't open %s", fname);
		MessageBox(NULL, ts, "Error", MB_ICONERROR | MB_OK);

		goto Error;
	}

	smafsize = GetFileSize(hf, NULL);

	if(!(smaf = halloc(smafsize)))
	{
		MessageBox(NULL, "Can't alloc memory", "Error", MB_ICONERROR | MB_OK);
		goto Error;
	}
	if(!ReadFile(hf, smaf, smafsize, &read, NULL))
	{
		read = 0;
	}
	if(read != smafsize)
	{
		MessageBox(NULL, "Couldn't read file.", "Error", MB_ICONERROR | MB_OK);
		goto Error;
	}

	CloseHandle(hf);

	{
		CHUNK chk;
		if(!findchunk4(smaf, smafsize, "MMMD", &chk))
		{
			MessageBox(NULL, "This is not SMAF", "Error", MB_ICONERROR | MB_OK);
			return;
		}

		chk.size -= 2;	// MMMD's CRC

		if(findchunk3(chk.data, chk.size, "MTR", &chk))
		{
			CHUNK dummy;

			switch(chk.data[2])
			{
			case 0x00: dbase = 1; break;
			case 0x01: dbase = 2; break;
			case 0x02: dbase = 4; break;
			case 0x03: dbase = 5; break;
			case 0x10: dbase = 10; break;
			case 0x11: dbase = 20; break;
			case 0x12: dbase = 40; break;
			case 0x13: dbase = 50; break;
			default: MessageBox(NULL, "Unknown duration timebase", "Warning", MB_ICONWARNING | MB_OK); break;
			}

			switch(chk.data[3])
			{
			case 0x00: gbase = 1; break;
			case 0x01: gbase = 2; break;
			case 0x02: gbase = 4; break;
			case 0x03: gbase = 5; break;
			case 0x10: gbase = 10; break;
			case 0x11: gbase = 20; break;
			case 0x12: gbase = 40; break;
			case 0x13: gbase = 50; break;
			default: MessageBox(NULL, "Unknown gatetime timebase", "Warning", MB_ICONWARNING | MB_OK); break;
			}

			if(chk.data[0] == 0x00)	// HPS
			{
				chk.data += 4 + 2;

				if(findchunk4(chk.data, chk.size, "Mtsu", &dummy))
					ConvertEvents2(dummy.data, dummy.size, TRUE);

				if(findchunk4(chk.data, chk.size, "Mtsq", &dummy))
					ConvertEvents2(dummy.data, dummy.size, FALSE);
			}

			if(chk.data[0] == 0x01 || chk.data[0] == 0x02)	// MSF
			{
				chk.data += 4 + 16;
				if(findchunk4(chk.data, chk.size, "Mtsu", &dummy))
					ConvertEvents(dummy.data, dummy.size, TRUE);

				if(findchunk4(chk.data, chk.size, "Mtsq", &dummy))
				{
					if(chk.data[-20] == 0x01)
					{
						BYTE* data;
						data = halloc(SwapDword(dummy.data));
						huffmanDecode(data, dummy.data + 4, SwapDword(dummy.data), dummy.size);
						ConvertEvents(data, SwapDword(dummy.data), FALSE);
						hfree(data);
					}
					if(chk.data[-20] == 0x02)
						ConvertEvents(dummy.data, dummy.size, FALSE);
				}
			}
		}
	}

	return;
Error:
	if(smaf)
	{
		hfree(smaf);
		smaf = NULL;
	}
	if(hf)
	{
		CloseHandle(hf);
	}

	return;
}

LRESULT CALLBACK MainWndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	static HMIDIOUT hmi;
	static BOOL sounding;
	static DWORD start;
	static BYTE velocity[16];

	switch(msg)
	{
	case WM_CREATE:
		midiOutOpen(&hmi, MIDIMAPPER, 0, 0, CALLBACK_NULL);
		midiOutShortMsg(hmi, MIDIMSG3(0x90, 60, 0x7F));
		midiOutShortMsg(hmi, MIDIMSG3(0x90, 64, 0x7F));
		midiOutShortMsg(hmi, MIDIMSG3(0x90, 67, 0x7F));

		{
			SCROLLINFO si;

			si.cbSize = sizeof(si);
			si.fMask = SIF_ALL | SIF_DISABLENOSCROLL;
			si.nMin = 0;
			si.nMax = 100;
			si.nPage = 10;
			si.nPos = 0;
			SetScrollInfo(hWnd, SB_HORZ, &si, TRUE);

			si.nMin = 0;
			si.nMax = 10000;
			si.nPage = 25;
			si.nPos = 0;
			SetScrollInfo(hWnd, SB_VERT, &si, TRUE);
		}

		break;
	case WM_MOUSEWHEEL:
		if(LOWORD(wParam) & MK_SHIFT)
		{
			if((short)HIWORD(wParam) > 0)
			{
				if(LOWORD(wParam) & MK_CONTROL)
					PostMessage(hWnd, WM_HSCROLL, SB_PAGELEFT, 0);
				else
					PostMessage(hWnd, WM_HSCROLL, SB_LINELEFT, 0);
			}else
			{
				if(LOWORD(wParam) & MK_CONTROL)
					PostMessage(hWnd, WM_HSCROLL, SB_PAGERIGHT, 0);
				else
					PostMessage(hWnd, WM_HSCROLL, SB_LINERIGHT, 0);
			}
		}else
		{
			if((short)HIWORD(wParam) > 0)
			{
				if(LOWORD(wParam) & MK_CONTROL)
					PostMessage(hWnd, WM_VSCROLL, SB_PAGEUP, 0);
				else
					PostMessage(hWnd, WM_VSCROLL, SB_LINEUP, 0);
			}else
			{
				if(LOWORD(wParam) & MK_CONTROL)
					PostMessage(hWnd, WM_VSCROLL, SB_PAGEDOWN, 0);
				else
					PostMessage(hWnd, WM_VSCROLL, SB_LINEDOWN, 0);
			}
		}
		break;
	case WM_KEYDOWN:
		switch((int)wParam)
		{
		case VK_DOWN:
			PostMessage(hWnd, WM_VSCROLL, SB_LINEDOWN, 0);
			break;
		case VK_UP:
			PostMessage(hWnd, WM_VSCROLL, SB_LINEUP, 0);
			break;
		case VK_LEFT:
			PostMessage(hWnd, WM_HSCROLL, SB_LINELEFT, 0);
			break;
		case VK_RIGHT:
			PostMessage(hWnd, WM_HSCROLL, SB_LINERIGHT, 0);
			break;
		case VK_NEXT:
			PostMessage(hWnd, WM_VSCROLL, SB_PAGEDOWN, 0);
			break;
		case VK_PRIOR:
			PostMessage(hWnd, WM_VSCROLL, SB_PAGEUP, 0);
			break;
		}
		break;
	case WM_HSCROLL:
	case WM_VSCROLL:
		{
			SCROLLINFO scr;

			scr.cbSize = sizeof(scr);
			scr.fMask = SIF_ALL;
			if(msg == WM_HSCROLL)
				GetScrollInfo(hWnd, SB_HORZ, &scr);
			if(msg == WM_VSCROLL)
				GetScrollInfo(hWnd, SB_VERT, &scr);

			switch(LOWORD(wParam))
			{
			//case SB_TOP:
			case SB_LEFT:
				scr.nPos = scr.nMin;
				break;
			//case SB_BOTTOM:
			case SB_RIGHT:
				scr.nPos = scr.nMax;
				break;
			//case SB_LINEUP:
			case SB_LINELEFT:
				if(scr.nMin < scr.nPos)
					scr.nPos--;
				break;
			//case SB_LINEDOWN:
			case SB_LINERIGHT:
				if(scr.nPos < scr.nMax - 1)
					scr.nPos++;
				break;
			//case SB_PAGEUP:
			case SB_PAGELEFT:
				scr.nPos -= scr.nPage;
				break;
			//case SB_PAGEDOWN:
			case SB_PAGERIGHT:
				scr.nPos += scr.nPage;
				break;
			case SB_THUMBPOSITION:	
				scr.nPos = HIWORD(wParam);
				break;
			}

			scr.fMask = SIF_POS;
			if(msg == WM_HSCROLL)
				SetScrollInfo(hWnd, SB_HORZ, &scr, TRUE);
			if(msg == WM_VSCROLL)
				SetScrollInfo(hWnd, SB_VERT, &scr, TRUE);

			InvalidateRect(hWnd, NULL, FALSE);
		}

		return 0;
	case WM_PAINT:
		{
			PAINTSTRUCT ps;
			HDC hdc;

			hdc = BeginPaint(hWnd, &ps);

			{
				RECT rc;

				GetClientRect(hWnd, &rc);
				draw(hWnd, hdc, rc.right, rc.bottom);
			}

			EndPaint(hWnd, &ps);
		}
		break;
	case WM_COMMAND:
		switch(HIWORD(wParam))
		{
		case 0:	// Command
		case 1:	// Accelerator
			switch(LOWORD(wParam))
			{
			case ID_OPENSMAF:
				{
					OPENFILENAME ofn = {0};
					TCHAR fname[MAX_PATH];

					ofn.lStructSize = sizeof(ofn);
					ofn.Flags = OFN_FILEMUSTEXIST;
					ofn.hInstance = GetModuleHandle(NULL);
					ofn.hwndOwner = hWnd;
					ofn.lpstrTitle = "Hello.";
					fname[0] = '\0';
					ofn.lpstrFile = fname;
					ofn.nMaxFile = sizeof(fname);

					if(GetOpenFileName(&ofn))
					{
						SendMessage(hWnd, WM_COMMAND, ID_CLOSE, 0);

						LoadSmaf(fname);

						InvalidateRect(hWnd, NULL, FALSE);
					}
				}
				break;
			case ID_CLOSE:
				if(sounding)
					SendMessage(hWnd, WM_COMMAND, ID_STOP, 0);
				FreeSmaf();
				InvalidateRect(hWnd, NULL, FALSE);
				break;
			case ID_EXIT:
				PostMessage(hWnd, WM_CLOSE, 0, 0);
				break;
			case ID_HELPINDEX:
				MessageBox(hWnd, "There are no help.", "sorry", MB_ICONERROR | MB_OK);
				break;
			case ID_ABOUT:
				DialogBoxParam(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_ABOUT), hWnd, MessageDlgWndProc, 0);
				break;
			case ID_PLAY:
				if(smaf && !sounding)
				{
					int i;

					for(i = 0; i < 16; i++)
						velocity[i] = 64;

					sounding = TRUE;
					start = GetTickCount();
					current = eventlist;

					SetTimer(hWnd, 0, 10, NULL);
				}
				break;
			case ID_STOP:
				if(smaf && sounding)
				{
					NOTEOFF* no = noteofflist;

					while(no)
					{
						NOTEOFF* nx = no->next;
						hfree(no);
						no = nx;
					}
					noteofflist = NULL;

					sounding = FALSE;
					KillTimer(hWnd, 0);
					midiOutReset(hmi);

					SetWindowText(hWnd, "mmf tool");
				}
				break;
			}
			break;
		}
		break;
	case WM_TIMER:
		if(sounding && current)
		{
			DWORD curtime = GetTickCount() - start;
			NOTEOFF* no = noteofflist;
			NOTEOFF* preno = NULL;

			{
				TCHAR ts[256];
				DWORD i;

				wsprintf(ts, "mmf tool %06dms ", curtime);

				for(i = 0; i < 16; i++)
				{
					NOTEOFF* no = noteofflist;
					while(no)
					{
						if(no->ch == i)
							break;
						no = no->next;
					}
					if(no)
						lstrcat(ts, "1:");
					else
						lstrcat(ts, "0:");
					wsprintf(ts, "%s%03d ", ts, velocity[i]);
				}

				SetWindowText(hWnd, ts);
			}

			// noteoff
			while(no)
			{
				if(no->end <= curtime)
				{
					NOTEOFF* nx;
					midiOutShortMsg(hmi, MIDIMSG3(0x80 | no->ch, no->note, 0x7F));
					if(preno)
						preno->next = no->next;
					else
						noteofflist = no->next;
					nx = no->next;
					hfree(no);
					no = nx;
				}else
				{
					preno = no;
					no = no->next;
				}
			}

			// events
			while(current && current->time <= curtime)
			{
				if((current->status & 0xF0) == 0x80 || (current->status & 0xF0) == 0x90)
				{
					NOTEOFF* no = noteofflist;

					if((current->status & 0xF0) == 0x90)
						velocity[current->status & 0x0F] = current->data[1];
					midiOutShortMsg(hmi, MIDIMSG3(0x90 | (current->status & 0x0F), current->data[0], velocity[current->status & 0x0F]));

					if(!no)
					{
						noteofflist = halloc(sizeof(NOTEOFF));
						no = noteofflist;
					}else
					{
						while(no->next)
							no = no->next;
						no->next = halloc(sizeof(NOTEOFF));
						no = no->next;
					}
					no->ch = current->status & 0x0F;
					no->note = current->data[0];
					if((current->status & 0xF0) == 0x90)
						no->end = current->time + GetVVal(current->data + 2) * gbase;
					else
						no->end = current->time + GetVVal(current->data + 1) * gbase;
					no->next = NULL;
				}
				if((current->status & 0xF0) == 0xB0)
					midiOutShortMsg(hmi, MIDIMSG3(0xB0 | (current->status & 0x0F), current->data[0], current->data[1]));
				if((current->status & 0xF0) == 0xC0)
					midiOutShortMsg(hmi, MIDIMSG2(0xC0 | (current->status & 0x0F), current->data[0]));
				if((current->status & 0xF0) == 0xE0)
					midiOutShortMsg(hmi, MIDIMSG3(0xE0 | (current->status & 0x0F), current->data[0], current->data[1]));
				current = current->next;
			}

			if(!current)	// end
			{
				SendMessage(hWnd, WM_COMMAND, ID_STOP, 0);
			}
		}
		break;
	case WM_CLOSE:
		SendMessage(hWnd, WM_COMMAND, ID_CLOSE, 0);

		midiOutReset(hmi);
		midiOutClose(hmi);

		break;
	case WM_DESTROY:
		PostQuitMessage(0);
		break;
	}

	return DefWindowProc(hWnd, msg, wParam, lParam);
}

BOOL initWnd(void)
{
	WNDCLASSEX wc = {0};
	
	wc.cbSize = sizeof(wc);
	wc.hbrBackground = (HBRUSH)(COLOR_3DFACE + 1);
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hInstance = GetModuleHandle(NULL);
	wc.hIcon = LoadIcon(wc.hInstance, MAKEINTRESOURCE(IDI_MAIN));
	wc.lpfnWndProc = MainWndProc;
	wc.lpszClassName = "MMFToolWC";
	wc.style = CS_HREDRAW | CS_VREDRAW;
	wc.lpszMenuName = MAKEINTRESOURCE(IDR_MAIN);

	if(!RegisterClassEx(&wc))
		return FALSE;

	// I don't like WS_OVERLAPPED.
	if(!CreateWindowEx(WS_EX_CLIENTEDGE, wc.lpszClassName, "mmf tool", WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX | WS_SIZEBOX | WS_MAXIMIZEBOX | WS_VISIBLE | WS_HSCROLL | WS_VSCROLL, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, NULL, NULL, wc.hInstance, NULL))
		return FALSE;

	return TRUE;
}

void WinMainCRTStartup(void)
{
	if(!initWnd())
		ExitProcess(0);

	{
		MSG msg;
		HACCEL at;

		at = LoadAccelerators(GetModuleHandle(NULL), MAKEINTRESOURCE(IDR_MAIN));

		while(GetMessage(&msg, NULL, 0, 0))
		{
			if(!TranslateAccelerator(msg.hwnd, at, &msg))
				TranslateMessage(&msg);
			DispatchMessage(&msg);
		}

		ExitProcess(msg.wParam);
	}
}
